# coding=utf-8

"""This file is part of PROMISE.

	PROMISE is free software: you can redistribute it and/or modify it
	under the terms of the GNU Lesser General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	PROMISE is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
	Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with PROMISE. If not, see
	<http://www.gnu.org/licenses/>.

Promise v1 was written by Romain Picot
Promise v2 has been written from v1 by Thibault Hilaire and Sara Hoseininasab
Promise v3 has been written from v2 by Thibault Hilaire and Fabienne JÉZÉQUEL and Xinye Chen
	Promise v3 enables version check, pip installing, arbitrary precision, etc., features.
	Sorbonne Université
	LIP6 (Computing Science Laboratory)
	Paris, France. Contact: thibault.hilaire@lip6.fr, fabienne.jezequel@lip6.fr, xinyechenai@gmail.com

"""


class PromiseError(Exception):
	"""Custom class to raise Promise Exception"""
	pass


class PromiseCompilationError(PromiseError):
	"""Custom class to raise Promise Exception when Compilation failed"""
	pass



# coding: utf-8

from setuptools import setup, find_packages


def readme():
    """Read the README file"""
    with open('README.rst', encoding='utf-8') as f:
        return f.read()


setup(
    name='cadnaPromise',
    version='0.3.9',
    description='Precision auto-tuning of floating-point variables in program',
    long_description=readme(),
    long_description_content_type='text/x-rst',

    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: C",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
    ],

    keywords='computer arithmetic mixed-precision precision auto-tuning',
    url='https://github.com/PEQUAN/cadnaPromise',

    author='LIP6 PEQUAN team',
    author_email='thibault.hilaire@lip6.fr; fabienne.jezequel@lip6.fr',

    license='GNU General Public License v3.0',

    packages=find_packages(),

    include_package_data=True,
    package_data={
        "cadnaPromise.promise": ["extra/*"],

        "cadnaPromise": [
            "deltadebug/*",
            "cadna/*",
            "cache/*",
        ],
    },

    install_requires=[
        'colorlog',
        'colorama',
        'tqdm',
        'regex',
        'pyyaml',
        'docopt-ng',
    ],

    tests_require=[
        'pytest',
        'pytest-cov',
    ],

    setup_requires=[
        'setuptools',
        'colorlog',
        'colorama',
        'regex',
    ],

    extras_require={
        'with_doc': [
            'sphinx',
            'sphinx_bootstrap_theme',
        ],
    },

    # ------------------------------------------------------------
    # entry points
    # ------------------------------------------------------------
    entry_points={
        'console_scripts': [
            'promise=cadnaPromise.run:runPromise',
            'activate-promise=cadnaPromise.install:activate',
            'deactivate-promise=cadnaPromise.install:deactivate',
            'load_CADNA_PATH=cadnaPromise.run:loadCADNA',
            'promise-batch=cadnaPromise.run:run_experiment_and_plot',
        ]
    },

    zip_safe=False,
)



# coding=utf-8

"""This file is part of PROMISE.

	PROMISE is free software: you can redistribute it and/or modify it
	under the terms of the GNU Lesser General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	PROMISE is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
	Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with PROMISE. If not, see
	<http://www.gnu.org/licenses/>.

Promise v1 was written by Romain Picot
Promise v2 has been written from v1 by Thibault Hilaire and Sara Hoseininasab
Promise v3 has been written from v2 by Thibault Hilaire, Fabienne JÉZÉQUEL and Xinye Chen
	Promise v3 enables version check, pip installing, arbitrary precision, etc., features.
	Sorbonne Université
	LIP6 (Computing Science Laboratory)
	Paris, France. Contact: thibault.hilaire@lip6.fr, fabienne.jezequel@lip6.fr, xinyechenai@gmail.com

	
© Thibault Hilaire and Fabienne JÉZÉQUEL, April 2024

	Some useful functions for the logger
	Mainly, each file instanciates a PrLogger (that contains an instance of a logger)
	At the beginning, the logger is configured (from the yaml file and the command line options)

	Two handlers handle the logs:
	- one for display the information to the terminal
	- one for the log file
	The level set depends on the verbosity (--verbosity and --verbosityLog options). There are 5 levels of verbosity

	List of logging level, and the verbosity level (for each level, display everything above):
	- (40) ERROR (error messages)
	- (25) MESSAGE (start up, end messages)             <- verbosity 0
	- (20) INFO (steps done)                            <- verbosity 1
	- (10) DEBUG (some debug info)                      <- verbosity 2
	- (7) OUTPUT (output of the runs)
	- (6) COMMAND (command that are done)               <- verbosity 3
	- (5) LOW_DEBUG (low level debug messages)
	- (3) OUTPUT_PROMISE (values display by promise.h)  <- verbosity 4



	© Thibault HILAIRE, April 2020
"""


from colorlog import ColoredFormatter  # logging with colors
import logging
from os.path import join, expanduser
from .errors import PromiseError
from tqdm import tqdm


# logging level (see logging module) and their color, with respect to the verbosity
levels = [25, 20, 10, 6, 3]
log_colors = {
	'ERROR': 'red', 'MESSAGE': 'blue', 'INFO': 'white', 'OUTPUT': 'cyan,', 'DEBUG': 'green',
	'COMMAND': 'purple', 'LOW_DEBUG': 'yellow', 'OUTPUT_PROMISE': 'cyan'
}


class TqdmLoggingHandler(logging.Handler):
	"""	A dedicated Logging Handler
	to be used with tqdm progress bar, see
https://stackoverflow.com/questions/38543506/change-logging-print-function-to-tqdm-write-so-logging-doesnt-interfere-wit
	for more informationns"""
	def __init__(self, level=logging.NOTSET):
		"""Set up the super logger"""
		super().__init__(level)

	def emit(self, record):
		"""emit method, display message using tqdm.write instead of print"""
		msg = self.format(record)
		tqdm.write(msg)
		self.flush()



class PrLogger:
	"""Dedicated logger for Promise
	dedicated levels, dedicated handlers that are removed at the end"""

	def __init__(self):
		"""just get the root logger"""
		self._logger = logging.getLogger('promise')
		self._step = 'a'  # step

	def reset(self):
		"""Reset the logger"""
		self._step = 'a'
		for h in self._logger.handlers[:]:
			h.close()
			self._logger.removeHandler(h)

		self._logger.handlers.clear() # add by Xinye

	# -- log levels --
	def error(self, msg, *args, **kws):
		"""log at message at level 40 (error)"""
		self._logger.log(40, msg, *args, **kws)

	def message(self, msg, *args, **kws):
		"""log at message at level 25 (message)"""
		self._logger.log(25, msg, *args, **kws)

	def info(self, msg, *args, **kws):
		"""log at message at level 20 (info)"""
		self._logger.log(20, msg, *args, **kws)

	def debug(self, msg, *args, **kws):
		"""log at message at level 10 (debug)"""
		self._logger.log(10, msg, *args, **kws)

	def output(self, msg, *args, **kws):
		"""log at message at level 5 (low level debug)"""
		self._logger.log(7, msg, *args, **kws)

	def command(self, msg, *args, **kws):
		"""log at message at level 5 (low level debug)"""
		self._logger.log(6, msg, *args, **kws)

	def lowdebug(self, msg, *args, **kws):
		"""log at message at level 5 (low level debug)"""
		self._logger.log(5, msg, *args, **kws)

	def outputPromise(self, msg, *args, **kws):
		"""log at message at level 5 (very low level debug)"""
		self._logger.log(3, msg, *args, **kws)

	def step(self, string):
		"""log a new step"""
		self._logger.info(self._step + ') ' + string)
		self._step = chr(ord(self._step) + 1)


	def configureLogger(self, options):
		"""Configure the logger from the options of the *command line* (only)
		(not from the configuration file, because we need to configure the logger
		before reading the configuration file, otherwise we cannot log the potential errors
		while parsing it)
		"""
		# get the path
		path = options['--path'] if options['--path'] else ''
		# get the verbosity
		verbosity = getInt(self._logger, options, 'verbosity', 1, 4, default=1)
		verbosityLog = getInt(self._logger, options, 'verbosityLog', 1, 4, default=verbosity)
		options['--verbosityLog'] = verbosityLog
		# set new logger levels for messages and low level debug
		logging.addLevelName(25, "MESSAGE")
		logging.addLevelName(7, "OUTPUT")
		logging.addLevelName(6, "COMMAND")
		logging.addLevelName(5, "LOW_DEBUG")
		logging.addLevelName(3, "OUTPUT_PROMISE")
		# stream the log to the console, with colors
		stream_handler = TqdmLoggingHandler()
		stream_handler.setLevel(levels[verbosity])
		stream_handler.setFormatter(ColoredFormatter("%(log_color)s%(message)s%(reset)s", log_colors=log_colors))
		self._logger.addHandler(stream_handler)
		# stream the log to the file
		if options['--log']:
			self.addFileLogger(path, options['--log'], verbosityLog)
		# set the logger to the minimum level (the handler will filter the messages)
		self._logger.setLevel(3)


	def addFileLogger(self, path, filename, verbosity):
		"""Add a file handler for the logger
		it is be called from `configureLogger` (when the command line options is parsed)
		and also by `getYMLoptions` when the yml file is parsed
		(the file logger should be configure *before* parsing the yml file,
		in order to log the potential errors of the parsing)"""
		file_handler = logging.FileHandler(expanduser(join(path, filename)), mode='w')
		file_handler.setLevel(levels[verbosity])
		error_formatter = logging.Formatter('%(asctime)s: %(message)s', "%m/%d %H:%M:%S")
		file_handler.setFormatter(error_formatter)
		self._logger.addHandler(file_handler)



def getInt(logger, options, name, mini, maxi, default=None):
	"""get an integer in the options, betwen min and max
	if not present, the default value is taken
	use the logger to log the errors, if happens"""
	optName = '--' + name
	try:
		val = int(options[optName]) if (optName in options) and (options[optName]) else default
	except ValueError as err:
		logger.error('The ' + name + ' is incorrect (`--' + name + '` option)')
		raise PromiseError(err) from err
	if not (mini <= val <= maxi):
		logger.error('The %s should be between %d and %d (included) (`--%s` option)' % (name, mini, maxi, name))
		raise PromiseError
	return val






# coding=utf-8
# This file is part of PROMISE.
#
# 	PROMISE is free software: you can redistribute it and/or modify it
# 	under the terms of the GNU Lesser General Public License as
# 	published by the Free Software Foundation, either version 3 of the
# 	License, or (at your option) any later version.
#
# 	PROMISE is distributed in the hope that it will be useful, but WITHOUT
# 	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# 	or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# 	Public License for more details.
#
# 	You should have received a copy of the GNU Lesser General Public
# 	License along with PROMISE. If not, see
# 	<http://www.gnu.org/licenses/>.
#
# Promise v1 was written by Romain Picot
# Promise v2 has been written from v1 by Thibault Hilaire and Sara Hoseininasab
# Promise v3 has been written from v2 by  Thibault Hilaire, Fabienne JÉZÉQUEL and Xinye Chen
#   Promise v3 enables version check, pip installing, arbitrary precision, etc., features.
# 	  Sorbonne Universitéx, LIP6 (Computing Science Laboratory), Paris, France. 
#     Contact: thibault.hilaire@lip6.fr, fabienne.jezequel@lip6.fr, xinyechenai@gmail.com
#
# 	contain the entry function, called to run Promise
#
# 	© Thibault Hilaire and Fabienne JÉZÉQUEL, April 2024


"""
activate-promise

Usage:
    activate-promise [[--CC=<compiler1>|--CXX=<compiler2>] | [--CC=<compiler1> --CXX=<compiler2>]]
    activate-promise (-h | --help)
    activate-promise (--version)
    deactivate-promise 
    deactivate-promise (-h | --help)
    deactivate-promise (--version)

Options:
	-h --help                     Show this screen.
    --version                     Show version.
    --CC=<compiler1>        	  Set compiler for C program [default: g++].
    --CXX=<compiler2>             Set compiler for C++ program [default: g++].
"""


import os
import sys
# from .utils import customArgParser, __params__
from docopt import docopt
from .utils import get_version

cachePath = "/cache"

# ./configure CXX=YourC++Compiler 
# ./configure --prefix=TheInstallationDirectory
# ./configure CXX=YourC++Compiler  --prefix=TheInstallationDirectory 
# ./configure --prefix=`pwd` --enable-fortran

# Macos 
# Brew gcc/7.3.0_1 (gcc, g++, gfortran)
# ./configure --prefix=`pwd` CC=gcc-7 CXX=g++-7 --enable-fortran
# ./configure --prefix=`pwd` CC=clang CXX=clang++ OPENMP_CFLAGS="-I/usr/local/Cellar/libomp/5.0.1/include -Xclang -fopenmp" OPENMP_CXXFLAGS="-I/usr/local/Cellar/libomp/5.0.1/include -Xclang -fopenmp" OPENMP_LDFLAGS="-L/usr/local/Cellar/libomp/5.0.1/lib -lomp"
# ./configure --prefix=`pwd` CC=clang CXX=clang++ OPENMP_CFLAGS="-I/usr/local/Cellar/libomp/7.0.0/include -Xclang -fopenmp" OPENMP_CXXFLAGS="-I/usr/local/Cellar/libomp/7.0.0/include -Xclang -fopenmp" OPENMP_LDFLAGS="-L/usr/local/Cellar/libomp/7.0.0/lib -lomp"
# On OSX gcc may be a wrapper for clang. In that case you can use:
# ./configure --prefix=`pwd` CC=clang CXX=clang++

def activate(argv=None): 
    
    import subprocess
    import platform

    opt_system = platform.system()
    curr_loc = os.path.dirname(os.path.realpath(__file__))

    install_cadna = True
    # args = customArgParser(sys.argv[1:] if argv == {} else argv, __params__)      # parse the command line

    args = docopt(__doc__, argv=sys.argv[1:] if argv is {} else argv, 
                  version=get_version(os.path.dirname(os.path.realpath(__file__))+'/__init__.py')
                  )

    compiler_CXX = args['--CXX']
    compiler_CC = args['--CC']
        
    if not os.path.exists(curr_loc + cachePath):
        os.makedirs(cachePath)
        
    if 'CADNA_PATH' in os.environ:  # check environmental variables
        import logging
        logging.basicConfig()
        log = logging.getLogger()

        log.warning("It looks like your machine has CADNA installed, are you sure to proceed CADNA installation? ")
        check_point = input("Please answer 'yes' or 'no':")

        if check_point.lower() in {'yes', 'y'}:
            install_cadna = True

        else:
            install_cadna = False

    curr_loc = os.path.dirname(os.path.realpath(__file__))

    if os.path.exists(curr_loc + cachePath):
        if os.path.isfile(curr_loc + cachePath + '/CXX.txt'):
            with open(curr_loc+cachePath+"/CXX.txt", "r") as file:
                compiler = file.read().replace('\n', '')
                print('check compilers:', compiler)

                if compiler_CXX != compiler:
                    install_cadna = True
                    

    os.chdir(curr_loc+'/cadna')

    if install_cadna: # install cadna if a.out and libcadnaC.a is not found
        with open(curr_loc + cachePath + "/CC.txt", "w") as file:
            file.write(compiler_CC)
            
        with open(curr_loc + cachePath + "/CXX.txt", "w") as file:
            file.write(compiler_CXX)

        # must set environmental variables
        if not os.path.isfile('a.out') and not os.path.isfile('lib/libcadnaC.a'):
            
            if opt_system in {'Linux', 'posix', 'Darwin'}:
                if compiler_CC == 'g++' and compiler_CXX == 'g++':
                    subprocess.call('bash run_unix.sh', shell=True)

                elif compiler_CC != 'g++' and compiler_CXX != 'g++':
                    subprocess.call('bash run_unix.sh '+compiler_CC+' '+compiler_CXX, shell=True)

                elif compiler_CC != 'g++' and compiler_CXX == 'g++':
                    subprocess.call('bash run_unix_cc.sh '+compiler_CC, shell=True)

                else:
                    subprocess.call('bash run_unix_cxx.sh '+compiler_CXX, shell=True)

    


def deactivate(): 
    import subprocess
    curr_loc = os.path.dirname(os.path.realpath(__file__))

    if not os.path.exists(curr_loc + cachePath):
        os.makedirs(cachePath)

    curr_loc = os.path.dirname(os.path.realpath(__file__))
    os.chdir(curr_loc+'/cadna')
    subprocess.call('make clean', shell=True)
    subprocess.call('bash clean.sh', shell=True)